// mtx_loader.c - (C) Victor Cerullo 2007
// version 1.4 - April 2007
//
// multiplatform code compatible with Apple Xcode and Cygwin gcc for Windows
// (same functionalities as mtx_loader version 1.3, with improved file parser)

#include <ext.h>
#include <ext_path.h>
#include <math.h>

typedef struct mtx_loader
{
    t_object p_ob;
    long p_value0;   // MIDI note value - received from left inlet
    void *p_outlet;  // frequency float outlet
    void *p_out1;    // list of MIDI cents (for sending to a coll object)
    void *p_out2;    // list of frequencies (for sending to a coll object)
    void *p_out3;    // rightmost outlet bangs in case of successful scale expansion
} t_mtx_loader;

double FTable[128];  // Frequency table (keymap)
t_symbol *ps_nothing;
void *mtx_loader_class;
void mtx_loader_expand(t_mtx_loader *x, short key, short period, short intervals);
void mtx_loader_read(t_mtx_loader *x, t_symbol *s);
void mtx_loader_doread(t_mtx_loader *x, t_symbol *s, short argc, t_atom *argv);
void mtx_loader_reset(t_mtx_loader *x);
void mtx_loader_dump(t_mtx_loader *x);
void mtx_loader_loadbang(t_mtx_loader *x);
void mtx_loader_sendlist(t_mtx_loader *x);
void mtx_loader_int(t_mtx_loader *x, long n);
void mtx_loader_list(t_mtx_loader *x, t_symbol *s, short ac, t_atom *av);
void mtx_loader_retune(t_mtx_loader *x, t_symbol *s, short ac, t_atom *av);
void mtx_loader_cent(t_mtx_loader *x, t_symbol *s, short ac, t_atom *av);
void mtx_loader_assist(t_mtx_loader *x, void *b, long m, long a, char *s);
void *mtx_loader_new(long n);

void main(void)
{
    setup((t_messlist **)&mtx_loader_class, (method)mtx_loader_new, 0L, (short)sizeof(t_mtx_loader), 0L, A_DEFLONG, 0);
    addint((method)mtx_loader_int);
    addmess((method)mtx_loader_read, "read", A_DEFSYM,0);   // open, read and process tuning file
    addmess((method)mtx_loader_reset, "reset", A_DEFSYM,0); // reset to 12-tET keymap
    addmess((method)mtx_loader_dump, "dump", A_DEFSYM,0);   // bang and dump frequency keymap
    addmess((method)mtx_loader_list,"list", A_GIMME,0);     // process list input (MIDI note n, frequency)
    addmess((method)mtx_loader_retune,"retune", A_GIMME,0); // retune keymap to a given reference frequency (MIDI note n, frequency)
    addmess((method)mtx_loader_cent,"cent", A_GIMME,0);     // transpose keymap (+/- cent)
    addmess((method)mtx_loader_assist, "assist", A_CANT, 0);
    addmess((method)mtx_loader_loadbang, "loadbang", A_CANT, 0);
    ps_nothing = gensym("");
}

void mtx_loader_expand(t_mtx_loader *x, short key, short period, short intervals)
{
    double FRatio[128];
    short i, count;
    short UpperLimit;
    if (intervals==0) {
        // case of ConsiderIntervals checkbox not marked
        for (i=key-1; i >= 0; --i) FTable[i]=FTable[i+period]/2;
        for (i=key+period; i < 128; ++i) FTable[i]=FTable[i-period]*2;
        mtx_loader_dump(x);
    }
    else {
        // case of ConsiderIntervals checkbox marked
        if (key+period-1 > 126) UpperLimit=126; else UpperLimit=key+period-1;
        for (i=key; i<=UpperLimit; ++i) {
            if (FTable[i]!=0) FRatio[i-key]=FTable[i+1]/(FTable[i]); else {
                if (FTable[i+1]==0) FRatio[i-key]=1; else FRatio[i-key]=999999999;
        }}
        if (period>1) {
            count=0;
            for (i=key+period; i<=127; ++i) {
                FTable[i]=FTable[i-1]*FRatio[count];
                ++count;
                if (count>period-2) count=0;
            }
            count=period-2;
            for (i=1; i<=key; ++i) {
                FTable[key-i]=FTable[key-i+1]/FRatio[count];
                --count;
                if (count<0) count=period-2;
            }
            mtx_loader_dump(x);
        }
        else {
            post("mtx_loader error: INTERVALS mode requires at least two tones");
    }}
}

void mtx_loader_read(t_mtx_loader *x, t_symbol *s) {
    defer(x,(method)mtx_loader_doread,s,0,0);
}

void mtx_loader_doread(t_mtx_loader *x, t_symbol *s, short argc, t_atom *argv)
{
    char filename[256], fname[256];
    filename[0] = 0;
    fname[0] = 0;
    t_filehandle f_fh;
    long size, bytecount;
    bytecount = 1;
    Byte data[16];
    short i, path, NoGoodFile;
    long outtype;
    const int MaxLineSize = 1024;
    char line[MaxLineSize];  // Single line read from file
    char c;
    short pos, keyline, freqline;
    int key;             // Scale start MIDI note
    int period;          // Number of tones the scale is composed of
    short intervals;     // "Consider Intervals" flag
    if (s==ps_nothing) {
        open_promptset("Select a Microtuner tuning file (.mtx)");
        if (open_dialog(filename, &path, &outtype, 0L, 0)) {
            post("mtx_loader warning: no file selected");
            return;
        }
        } else {
        strcpy(fname, s->s_name);
        strcpy(filename, s->s_name);
        if (locatefile_extended(fname, &path, &outtype, 0L, 0)) {
            post("mtx_loader error: file not found");
            return;
        }
    }
    if (!path_opensysfile(filename,path,&f_fh,READ_PERM)) {
        post("mtx_loader: parsing file %s", filename);
        intervals = 1;
        period = 0;
        key = 0;
        keyline = 0;
        freqline = 0;
        sysfile_geteof(f_fh,&size);
        NoGoodFile = 0;
        do {
            pos = 0;
            do {
                c = '\0';
                if (sysfile_read(f_fh,&bytecount,&data)) {
                    c = EOF;
                    break;
                }
                --size;
                c = (char)data[0];
                if ((c=='\0')||(c=='\r')||(c=='\n')||(c==EOF)||(size==0)) break;
                if ((pos==0)&&(keyline!=1)&&(isdigit(c)||(c=='.'))) freqline = 1;
                if ((pos==0)&&(c=='@')&&(keyline==0)) keyline = 1; else line[pos++] = c;
            } while ((c != EOF) && (size > 0) && (pos < MaxLineSize));
            if (pos < MaxLineSize) {
                line[pos] = '\0';
                if (strcmp(line,":absolute")==0) intervals=0;
                if (keyline==1) {
                    key = atoi(line);
                    if ((key>=0) && (key<128)){
                        post("mtx_loader: scale starts at MIDI key %d", key);
                        for (i=0; i < 128; ++i) FTable[i]=0;
			keyline = 99;
                    } else NoGoodFile = 1;
                }
                if ((keyline==99)&&(freqline==1)) {
                    if ((key+period) < 128) {
                        FTable[key+period] = atof(line);
                        ++period;
                        freqline = 0;
                    } else NoGoodFile = 1;
                }
            } else NoGoodFile = 1;
        } while ((c != EOF) && (size > 0) && (!NoGoodFile));
        sysfile_close(f_fh);
        if ((keyline!=99)||(period==0)) NoGoodFile = 1;
	if (!NoGoodFile) {
            post("mtx_loader: number of tones in the scale = %d", period);
            if (!intervals) post("mtx_loader: expansion mode = ABSOLUTE");
            else post("mtx_loader: expansion mode = INTERVALS");
            mtx_loader_expand(x, key,period,intervals);   
        } else post("mtx_loader error: file %s is not a valid mtx tuning file", &filename);
    }  else post("mtx_loader error: unable to open file %s", &filename);
}

void mtx_loader_reset(t_mtx_loader *x)
{
    // Initialize 12-tET default scale seed and expand keymap
    short i;
    for (i=0; i < 128; ++i) FTable[i]=0;
    FTable[69]=440;
    FTable[70]=466.16376151809;
    mtx_loader_expand(x, 69,2,1);
}

void mtx_loader_dump(t_mtx_loader *x)
{
    outlet_bang (x->p_out3);
    mtx_loader_sendlist(x);
}

void mtx_loader_sendlist(t_mtx_loader *x)
{
    t_atom FreqList[2], CentList[2];
    short i;
    double ET;
    for (i=0; i < 128; ++i) {
        SETLONG(CentList, i);
        ET = 440 * pow(2, (((double)i-69)/12));
        SETFLOAT(CentList+1, (100*i+(log(FTable[i])-log(ET))*1200/log(2)));
        SETLONG(FreqList, i);
        SETFLOAT(FreqList+1, FTable[i]);
        outlet_list(x->p_out1, 0L, 2, CentList);
        outlet_list(x->p_out2, 0L, 2, FreqList);
    }
}

void mtx_loader_int(t_mtx_loader *x, long n)
{
    float frequency;
    x->p_value0 = n;
    if (n<0) {
        x->p_value0 = 0;
    };
    if (n>127) {
        x->p_value0 = 127;
    };
    frequency = FTable[x->p_value0];
    outlet_float(x->p_outlet, frequency);
}

void mtx_loader_list(t_mtx_loader *x, t_symbol *s, short ac, t_atom *av)
{
    short i;
    if (ac==2) {
        if (av->a_type==A_LONG) {
            if ((av->a_w.w_long >= 0)&&(av->a_w.w_long < 128)) {
                i = av->a_w.w_long;
                av++;
                if (av->a_type==A_FLOAT) {
                    if (av->a_w.w_float >= 0) {
                        FTable[i]=(double)av->a_w.w_float;
                        mtx_loader_dump(x);
                    }
                }
                else if (av->a_type==A_LONG) {
                    if (av->a_w.w_long >= 0) {
                        FTable[i]=(double)av->a_w.w_long;
                        mtx_loader_dump(x);
                    }
                }
            }
        }
    }
}

void mtx_loader_retune(t_mtx_loader *x, t_symbol *s, short ac, t_atom *av)
{
    //retune only works when the frequency of the target MIDI note is > 0
    short i,j;
    double newref;
    if (ac==2) {
        if (av->a_type==A_LONG) {
            if ((av->a_w.w_long >= 0)&&(av->a_w.w_long < 128)) {
                i = av->a_w.w_long;
                av++;
                if (av->a_type==A_FLOAT) {
                    newref = (double)av->a_w.w_float;
                    if ((newref > 0)&&(FTable[i] > 0)) {
                        for (j=0; j < 128; ++j) FTable[j]=newref*(FTable[j]/FTable[i]);
                        mtx_loader_dump(x);
                    }
                }
                else if (av->a_type==A_LONG) {
                    newref = (double)av->a_w.w_long;
                    if ((newref > 0)&&(FTable[i] > 0)) {
                        for (j=0; j < 128; ++j) FTable[j]=newref*(FTable[j]/FTable[i]);
                        mtx_loader_dump(x);
                    }
                }
            }
        }
    }
}

void mtx_loader_cent(t_mtx_loader *x, t_symbol *s, short ac, t_atom *av)
{
    short i;
    if (ac==1) {
        if (av->a_type==A_LONG) {
            for (i=0; i < 128; ++i) FTable[i]=FTable[i]*pow(2, (double)av->a_w.w_long/1200);
            mtx_loader_dump(x);
        }
        else if (av->a_type==A_FLOAT) {
            for (i=0; i < 128; ++i) FTable[i]=FTable[i]*pow(2, (double)av->a_w.w_float/1200);
            mtx_loader_dump(x);
        }
    }
}

void mtx_loader_assist(t_mtx_loader *x, void *b, long m, long a, char *s)
{
    if (m == ASSIST_INLET)
    sprintf(s,"MIDI note in");
    else {
        switch (a) {
            case 0:
            sprintf(s,"frequency out (float)");
            break;
            case 1:
            sprintf(s,"MIDI cent keymap dump (list)");
            break;
            case 2:
            sprintf(s,"frequency keymap dump (list)");
            break;
            case 3:
            sprintf(s,"bang after scale expansion");
            break;
        }
    }
}

void mtx_loader_loadbang(t_mtx_loader *x)
{
    mtx_loader_dump(x);
}

void *mtx_loader_new(long n)
{
    t_mtx_loader *x;
    x = (t_mtx_loader *)newobject(mtx_loader_class);
    x->p_out3 = outlet_new(x,0);
    x->p_out2 = listout(x);
    x->p_out1 = listout(x);
    x->p_outlet = floatout(x);
    x->p_value0    = 0;
    mtx_loader_reset(x); // always start in 12-tET mode
    return(x);
}
